-----------------------------------------------------------------------------------------------
-- Client Lua Script for WhatTheRune
-- Copyright (c) CMBSolutions. All rights reserved
-- Distributed under the terms of the GPL v3 License.
--
-- Author: bMaurice bmaurice@cmbsolutions.nl
-----------------------------------------------------------------------------------------------

require "Window"
require "CraftingLib"
require "GameLib"
require "Item"
require "Tooltip"
require "Unit"

local knLevelRequirement = 15
local knAddonVersion = 2.2

local tHooks = 
{
	["Version"] = 1,
	["ViewRune"] = 2,
	["CraftRune"] = 4,
	["Equip"] = 8,
}

local tHookFunctions = {"WhatTheVersion", "OnExternalViewRune", "OnExternalCraftRune"}

local knAddonCompatibilities =
{
	["RuneMaster"]			=	{ nVersion = 0.4,	nMaxHook = tHooks["CraftRune"] },
	["WhatTheRuneTester"]	=	{ nVersion = 1,		nMaxHook = tHooks["Equip"] },
	["Unknown"]				=	{ nVersion = 0,		nMaxHook = tHooks["Version"] },
}

local karSigilElementsToSprite =
{
		[Item.CodeEnumSigilType.Air]		=	{ strName = Apollo.GetString("CRB_Air"),		strBright = "sprRunecrafting_Air",		strFade = "sprRunecrafting_AirFade",		strBright2 = "wtrSprites:wtrElementAirBright",		strFade2 = "wtrSprites:wtrElementAir"	},
		[Item.CodeEnumSigilType.Fire]		=	{ strName = Apollo.GetString("CRB_Fire"),		strBright = "sprRunecrafting_Fire",		strFade = "sprRunecrafting_FireFade",		strBright2 = "wtrSprites:wtrElementFireBright" ,	strFade2 = "wtrSprites:wtrElementFire"	},
		[Item.CodeEnumSigilType.Water]		=	{ strName = Apollo.GetString("CRB_Water"),		strBright = "sprRunecrafting_Water",	strFade = "sprRunecrafting_WaterFade",		strBright2 = "wtrSprites:wtrElementWaterBright" ,	strFade2 = "wtrSprites:wtrElementWater"	},
		[Item.CodeEnumSigilType.Earth]		=	{ strName = Apollo.GetString("CRB_Earth"),		strBright = "sprRunecrafting_Earth",	strFade = "sprRunecrafting_EarthFade",		strBright2 = "wtrSprites:wtrElementEarthBright" ,	strFade2 = "wtrSprites:wtrElementEarth"	},
		[Item.CodeEnumSigilType.Logic]		=	{ strName = Apollo.GetString("CRB_Logic"),		strBright = "sprRunecrafting_Logic",	strFade = "sprRunecrafting_LogicFade",		strBright2 = "wtrSprites:wtrElementLogicBright" ,	strFade2 = "wtrSprites:wtrElementLogic"	},
		[Item.CodeEnumSigilType.Life]		=	{ strName = Apollo.GetString("CRB_Life"),		strBright = "sprRunecrafting_Life",		strFade = "sprRunecrafting_LifeFade",		strBright2 = "wtrSprites:wtrElementLifeBright" ,	strFade2 = "wtrSprites:wtrElementLife"	},
		[Item.CodeEnumSigilType.Fusion]		=	{ strName = Apollo.GetString("CRB_Fusion"),		strBright = "sprRunecrafting_Fusion",	strFade = "sprRunecrafting_FusionFade",		strBright2 = "wtrSprites:wtrElementFusionBright" ,	strFade2 = "wtrSprites:wtrElementFusion"	},
		[Item.CodeEnumSigilType.Omni]		=	{ strName = Apollo.GetString("CRB_Omni"),		strBright = "sprRunecrafting_Omni",		strFade = "sprRunecrafting_OmniFade",		strBright2 = "wtrSprites:wtrElementOmniBright",		strFade2 = "wtrSprites:wtrElementOmni"	},	
}

local karSigilElementsToId =
{
	[Item.CodeEnumSigilType.Fire]		= 131,
	[Item.CodeEnumSigilType.Water]		= 132,
	[Item.CodeEnumSigilType.Earth]		= 133,
	[Item.CodeEnumSigilType.Air]		= 134,
	[Item.CodeEnumSigilType.Life]		= 135,
	[Item.CodeEnumSigilType.Logic]		= 136,
	[Item.CodeEnumSigilType.Fusion]		= 137,
	[Item.CodeEnumSigilType.Omni]		= 0,
}

local ktAttributeToText =
{
	[Unit.CodeEnumProperties.Dexterity] 					= Apollo.GetString("CRB_Finesse"),
	[Unit.CodeEnumProperties.Technology] 					= Apollo.GetString("CRB_Tech_Attribute"),
	[Unit.CodeEnumProperties.Magic] 						= Apollo.GetString("CRB_Moxie"),
	[Unit.CodeEnumProperties.Wisdom] 						= Apollo.GetString("UnitPropertyInsight"),
	[Unit.CodeEnumProperties.Stamina] 						= Apollo.GetString("CRB_Grit"),
	[Unit.CodeEnumProperties.Strength] 						= Apollo.GetString("CRB_Brutality"),

	[Unit.CodeEnumProperties.Armor] 						= Apollo.GetString("CRB_Armor") ,
	[Unit.CodeEnumProperties.ShieldCapacityMax] 			= Apollo.GetString("CBCrafting_Shields"),

	[Unit.CodeEnumProperties.AssaultPower] 					= Apollo.GetString("CRB_Assault_Power"),
	[Unit.CodeEnumProperties.SupportPower] 					= Apollo.GetString("CRB_Support_Power"),
	[Unit.CodeEnumProperties.Rating_AvoidReduce] 			= Apollo.GetString("CRB_Strikethrough_Rating"),
	[Unit.CodeEnumProperties.Rating_CritChanceIncrease] 	= Apollo.GetString("CRB_Critical_Chance"),
	[Unit.CodeEnumProperties.RatingCritSeverityIncrease] 	= Apollo.GetString("CRB_Critical_Severity"),
	[Unit.CodeEnumProperties.Rating_AvoidIncrease] 			= Apollo.GetString("CRB_Deflect_Rating"),
	[Unit.CodeEnumProperties.Rating_CritChanceDecrease] 	= Apollo.GetString("CRB_Deflect_Critical_Hit_Rating"),
	[Unit.CodeEnumProperties.ManaPerFiveSeconds] 			= Apollo.GetString("CRB_Attribute_Recovery_Rating"),
	[Unit.CodeEnumProperties.HealthRegenMultiplier] 		= Apollo.GetString("CRB_Health_Regen_Factor"),
	[Unit.CodeEnumProperties.BaseHealth] 					= Apollo.GetString("CRB_Health_Max"),

	[Unit.CodeEnumProperties.ResistTech] 					= Apollo.GetString("Tooltip_ResistTech"),
	[Unit.CodeEnumProperties.ResistMagic]					= Apollo.GetString("Tooltip_ResistMagic"),
	[Unit.CodeEnumProperties.ResistPhysical]				= Apollo.GetString("Tooltip_ResistPhysical"),

	[Unit.CodeEnumProperties.PvPOffensiveRating] 			= Apollo.GetString("Tooltip_PvPOffense"),
	[Unit.CodeEnumProperties.PvPDefensiveRating]			= Apollo.GetString("Tooltip_PvPDefense"),
}

local ktAttributeFromId =
{
	[1]		=	Unit.CodeEnumProperties.Dexterity,
	[2]		=	Unit.CodeEnumProperties.Technology,
	[3]		=	Unit.CodeEnumProperties.Magic,
	[4]		=	Unit.CodeEnumProperties.Wisdom,
	[5]		=	Unit.CodeEnumProperties.Stamina,
	[6]		=	Unit.CodeEnumProperties.Strength,
	[7]		=	Unit.CodeEnumProperties.Armor,
	[8]		=	Unit.CodeEnumProperties.AssaultPower,
	[9]		=	Unit.CodeEnumProperties.SupportPower,
	[10]	=	Unit.CodeEnumProperties.Rating_AvoidReduce,
	[11]	=	Unit.CodeEnumProperties.Rating_CritChanceIncrease,
	[12]	=	Unit.CodeEnumProperties.RatingCritSeverityIncrease,
	[13]	=	Unit.CodeEnumProperties.Rating_AvoidIncrease,
	[14]	=	Unit.CodeEnumProperties.PvPOffensiveRating,
	[15]	=	Unit.CodeEnumProperties.PvPDefensiveRating,

}

local knSaveVersion = 3

-----------------------------------------------------------------------------------------------
-- WhatTheRune Module Definition
-----------------------------------------------------------------------------------------------
local WhatTheRune = {}
 
local MissingMaterialsTooltipForm
local isAtEngravingStation = false
local isDebugging = false
local Rover
local md5

local externalAddon = "Unknown"
local isCalledExternal = false
local externalElement
local externalTypeTier
local externalSchematicId


-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function WhatTheRune:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self

    self.tUserSettings = {}
    self.tUserSettings.Version = false
    self.tUserSettings.DoNotChangePosition = true
	self.tUserSettings.FiltersShown = false
	self.tUserSettings.Craftable = false
	self.tUserSettings.Usable = false
	self.tUserSettings.NoGroups = false
	self.tUserSettings.Buff = false
	self.tUserSettings.LastKnownEngravingStationId = false
	
    if isDebugging then
        Rover = Apollo.GetAddon("Rover")
    end

    return o
end

function WhatTheRune:Init()
    local bHasConfigureFunction = true -- TODO: Figure out a way to save uservars without a configscreen
    local strConfigureButtonText = "WhatTheRune"
    local tDependencies = {
		
    }
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

-----------------------------------------------------------------------------------------------
-- Save & Restore functions to save or restore between session variables
-----------------------------------------------------------------------------------------------
function WhatTheRune:OnSave(eType)
    if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return
    end
	
	self.tUserSettings.Version = knAddonVersion

    local tSavedData = self.tUserSettings
    tSavedData.nOL, tSavedData.nOT, tSavedData.nOR, tSavedData.nOB = self.wndMain:GetAnchorOffsets()

    return tSavedData
end

function WhatTheRune:OnRestore(eType, tSavedData)
    if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return
    end

    if tSavedData ~= nil then
        for idx, item in pairs(tSavedData) do
            self.tUserSettings[idx] = item
        end
    end
	self.tUserSettings.Craftable = false
	self.tUserSettings.Usable = false
	self.tUserSettings.NoGroups= false
	self.tUserSettings.Buff = false

	if self.tUserSettings.Version == false or self.tUserSettings.Version ~= knAddonVersion then
		self.tUserSettings.DoNotChangePosition = true
		self.tUserSettings.Version = knAddonVersion
		self.tUserSettings.FiltersShown = false
	end
end

-----------------------------------------------------------------------------------------------
-- WhatTheRune OnLoad
-----------------------------------------------------------------------------------------------
function WhatTheRune:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("WhatTheRune.xml")
	self.xmlDoc:RegisterCallback("OnDocumentReady", self)
	Apollo.LoadSprites("wtrSprites.xml")
	Apollo.LoadSprites("wtrDialogSprites.xml")

end

-----------------------------------------------------------------------------------------------
-- WhatTheRune OnDocumentReady
-----------------------------------------------------------------------------------------------
function WhatTheRune:OnDocumentReady()
	if self.xmlDoc == nil then
		Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
		return
	end
		
	Apollo.RegisterSlashCommand("wtr", "OnWhatTheRuneOn", self)
    Apollo.RegisterSlashCommand("whattherune", "OnWhatTheRuneOn", self)

	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
	Apollo.RegisterEventHandler("AddonEvent_ToggleWhatTheRune", "OnToggleWhatTheRune", self)

	Apollo.RegisterEventHandler("GenericEvent_CraftingResume_OpenEngraving", "OnGenericEvent_CraftingResume_OpenEngraving", self)
	Apollo.RegisterEventHandler("GenericEvent_CraftingResume_CloseEngraving", "OnClose", self)
	Apollo.RegisterEventHandler("TradeskillEngravingStationClose", "OnClose", self)

	-- Unit events, when these trigger check if the unit is the engraving station
	Apollo.RegisterEventHandler("InteractiveUnitChanged", "OnInteractiveUnitChanged", self)
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)

	
	Apollo.RegisterEventHandler("UpdateInventory", "OnUpdateInventory", self)
	Apollo.RegisterEventHandler("PlayerEquippedItemChanged", "RedrawAll", self)
	Apollo.RegisterEventHandler("ItemModified", "RedrawAll", self)

	Apollo.RegisterTimerHandler("WhatTheRune_InitializeDelay", "RedrawAll", self)
	Apollo.CreateTimer("WhatTheRune_InitializeDelay", 1, false)
	Apollo.StopTimer("WhatTheRune_InitializeDelay")

	Apollo.RegisterTimerHandler("WhatTheRune_ConfirmPopupDelay", "OnWhatTheRune_ConfirmPopupDelay", self)
	Apollo.CreateTimer("WhatTheRune_ConfirmPopupDelay", 0.15, false)
	Apollo.StopTimer("WhatTheRune_ConfirmPopupDelay")

	Apollo.RegisterEventHandler("DragDropSysBegin", "OnSystemBeginDragDrop", self)
	Apollo.RegisterEventHandler("DragDropSysEnd", "OnSystemEndDragDrop", self)
end

-----------------------------------------------------------------------------------------------
-- WhatTheRune Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

function WhatTheRune:OnInterfaceMenuListHasLoaded()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "WhatTheRune", {"AddonEvent_ToggleWhatTheRune", "", "wtrSprites:wtrElementFireBright"})
end

function WhatTheRune:OnToggleWhatTheRune()
	isAtEngravingStation = false
    isCalledExternal = false
	self:OnWhatTheRuneOn()
end

function WhatTheRune:Initialize()
	if self.wndMain and self.wndMain:IsValid() then
		return
	end
	
	self.wndMain = Apollo.LoadForm(self.xmlDoc, "WhatTheRuneForm", nil, self)
	--self.wndWindowTemplate = Apollo.LoadForm(self.xmlDoc, "wtrMetal_ScrollList_Small", nil, self)
	local wndLst = self.wndMain:FindChild("RuneOptionBuffList"):SetWindowSubclass(self.wndWindowTemplate)
	
    if not self.tUserSettings.DoNotChangePosition and self.tUserSettings.nOL and self.tUserSettings.nOR >= 450 then
        self.wndMain:SetAnchorOffsets(self.tUserSettings.nOL, self.tUserSettings.nOT, self.tUserSettings.nOR, self.tUserSettings.nOB)
    end

    self.tUserSettings.DoNotChangePosition = false

	self.wndMain:FindChild("BGFrame"):FindChild("ShowTutorialsBtn"):Enable(false)
    self.wndMain:FindChild("BGFrame"):FindChild("ReloadButton"):Show(isDebugging)
	
	self.wndMain:FindChild("RuneCreationContainer"):FindChild("RuneCreationCraftBtn"):Enable(false)
	self.wndMain:FindChild("RuneCreationContainer"):FindChild("RuneCreationCraftBtn"):Show(self:HelperWhenAtStation(true, false))

	self.wndMain:FindChild("ToggleRuneCreation"):AttachWindow(self.wndMain:FindChild("RuneCreationContainer"))
	self.wndMain:FindChild("ToggleEquipRunes"):AttachWindow(self.wndMain:FindChild("EquipRunesContainer"))
	self.wndMain:FindChild("ToggleRuneOption"):AttachWindow(self.wndMain:FindChild("RuneOptionsContainer"))

	self.wndMain:FindChild("ToggleRuneOption"):Enable(true)
	self.wndMain:FindChild("ToggleRuneOption"):SetCheck(self.tUserSettings.FiltersShown)
	self.wndMain:FindChild("RuneOptionsContainer"):Show(self.wndMain:FindChild("ToggleRuneOption"):IsChecked())
	self.wndMain:FindChild("RuneOptionFilterCraftableBtn"):SetCheck(self.tUserSettings.Craftable)
	self.wndMain:FindChild("RuneOptionFilterUsableBtn"):SetCheck(self.tUserSettings.Usable)
	self.wndMain:FindChild("RuneOptionFilterNoGroupsBtn"):SetCheck(self.tUserSettings.NoGroups)
	self.wndMain:FindChild("RuneOptionFilterBuffBtn"):SetText(self:HelperIIF(self.tUserSettings.Buff, ktAttributeToText[ktAttributeFromId[self.tUserSettings.Buff]] , "--"))

	
	local wndMeasure = Apollo.LoadForm(self.xmlDoc, "EquipmentItem", nil, self)
	self.knEquipmentItemHeight = wndMeasure:GetHeight()
	wndMeasure:Destroy()

	wndMeasure = Apollo.LoadForm(self.xmlDoc, "TypeTierItem", nil, self)
	self.knDefaultTypeTierItemHeight = wndMeasure:GetHeight()
	wndMeasure:Destroy()

	self.bAllowClicks = true
	self.bDragging = false
	self.itemDragging = nil
	self.wndCurrentConfirmPopup = nil

	local bRequirementMet = true
	local nPlayerLevel = self:HelperWhenAtStation(GameLib.GetPlayerLevel(), 15)

	if nPlayerLevel < knLevelRequirement then
		self.wndMain:FindChild("RuneCreationItemList"):SetTextColor(ApolloColor.new("UI_WindowTextCraftingRedCapacitor"))
		self.wndMain:FindChild("RuneCreationItemList"):SetText(String_GetWeaselString(Apollo.GetString("Runecrafting_LevelRequirementNotMet"),  knLevelRequirement))
		bRequirementMet = false
	else
		self.wndMain:FindChild("RuneCreationItemList"):SetTextColor(ApolloColor.new("UI_TextMetalGoldHighlight"))
		self.wndMain:FindChild("RuneCreationItemList"):SetText("Select a rune element at the top to view the runes.")
	end

	local wndParent = self.wndMain:FindChild("RuneCreationElementList")
	for idx, eElement in pairs(Item.CodeEnumSigilType) do
		if eElement ~= Item.CodeEnumSigilType.Omni then
			local wndCurr = self:LoadByName("RuneCreationElementItem", wndParent, idx)
			wndCurr:FindChild("RuneCreationElementBtn"):SetData(eElement)
			wndCurr:FindChild("RuneCreationElementBtn"):Enable(bRequirementMet)
			wndCurr:FindChild("RuneCreationElementBtn"):SetTooltip(karSigilElementsToSprite[eElement].strName)
			wndCurr:FindChild("RuneCreationElementIcon"):SetSprite(karSigilElementsToSprite[eElement].strFade2)
			wndCurr:FindChild("RuneCreationElementName"):SetText(karSigilElementsToSprite[eElement].strName)
			wndCurr:FindChild("RuneCreationElementName"):SetTextColor("WindowTitleColor")
		end
	end
	wndParent:ArrangeChildrenHorz(0, function(a,b) return a:GetName() < b:GetName() end)
end

-----------------------------------------------------------------------------------------------
-- WhatTheRuneForm Functions
-----------------------------------------------------------------------------------------------
function WhatTheRune:OnClose( wndHandler, wndControl )
	if self.wndMain and self.wndMain:IsValid() then
        self.tUserSettings.nOL, self.tUserSettings.nOT, self.tUserSettings.nOR, self.tUserSettings.nOB = self.wndMain:GetAnchorOffsets()
		self.tUserSettings.FiltersShown = self.wndMain:FindChild("ToggleRuneOption"):IsChecked()
		self.wndMain:Destroy()
		self.wndMain = nil
	end

	Event_CancelEngravingStation()
	isAtEngravingStation = false
    isCalledExternal = false
	externalAddon = "Unknown"
end

function WhatTheRune:OnWhatTheRuneOn()
	self:Initialize()
	self:RedrawAll()
	self.wndMain:Invoke()
end

function WhatTheRune:OnGenericEvent_CraftingResume_OpenEngraving()
	self:Initialize()
	Apollo.StartTimer("WhatTheRune_InitializeDelay") -- Will do RedrawAll()
	isAtEngravingStation = true
	isCalledExternal = false
end

function WhatTheRune:OnRedrawAllFromUI(wndHandler, wndControl)
	self:RedrawAll()
end

function WhatTheRune:OnRuneCreationElementToggle(wndHandler, wndControl)
	for idx, wndCurr in pairs(self.wndMain:FindChild("RuneCreationElementList"):GetChildren()) do
		if wndCurr:FindChild("RuneCreationElementName") then
			wndCurr:FindChild("RuneCreationElementName"):SetTextColor(wndCurr:FindChild("RuneCreationElementBtn"):IsChecked() and "UI_WindowTitleYellow" or "WindowTitleColor")
		end
	end

	self.wndMain:FindChild("RuneCreationItemList"):SetText("Select a rune element at the top to view the runes.")
	self.wndMain:FindChild("RuneCreationItemList"):DestroyChildren()
	self.wndMain:FindChild("RuneCreationItemList"):RecalculateContentExtents()
	self:ResetRuneCreationBottom()
	self:RedrawAll()
end

function WhatTheRune:OnReload( wndHandler, wndControl, eMouseButton )
	RequestReloadUI()
end

function WhatTheRune:OnToggleRuneOptions( wndHandler, wndControl, eMouseButton )
	if not self.wndMain or not self.wndMain:IsValid() or not self.wndMain:IsVisible() then
		return
	end
	
	self.tUserSettings.FiltersShown = self.wndMain:FindChild("ToggleRuneOption"):IsChecked()
end

function WhatTheRune:OnOpenRuneMaster( wndHandler, wndControl, eMouseButton )
	local Runemaster = Apollo.GetAddon("RuneMaster")
	
	if Runemaster ~= nil then
		haveRunemaster = true
		Event_FireGenericEvent("RuneMaster_Show")
	else
		haveRunemaster = false
	end
end

function WhatTheRune:OnOptionFilterApply( wndHandler, wndControl, eMouseButton )
	self.tUserSettings.Craftable = wndHandler:GetParent():FindChild("RuneOptionFilterCraftableContainer"):FindChild("RuneOptionFilterCraftableBtn"):IsChecked()
	self.tUserSettings.Usable = wndHandler:GetParent():FindChild("RuneOptionFilterUsableContainer"):FindChild("RuneOptionFilterUsableBtn"):IsChecked()
	self.tUserSettings.NoGroups= wndHandler:GetParent():FindChild("RuneOptionFilterNoGroupsContainer"):FindChild("RuneOptionFilterNoGroupsBtn"):IsChecked()

	
	self.wndMain:FindChild("RuneCreationItemList"):SetText("Select a rune element at the top to view the runes.")
	self.wndMain:FindChild("RuneCreationItemList"):DestroyChildren()
	self.wndMain:FindChild("RuneCreationItemList"):RecalculateContentExtents()
	self:ResetRuneCreationBottom()
	self:RedrawAll()
end

function WhatTheRune:OnOptionFilterClear( wndHandler, wndControl, eMouseButton )
	self.tUserSettings.Craftable = false
	self.tUserSettings.Usable = false
	self.tUserSettings.NoGroups = false
	self.tUserSettings.Buff = false
	
	wndHandler:GetParent():FindChild("RuneOptionFilterCraftableContainer"):FindChild("RuneOptionFilterCraftableBtn"):SetCheck(false)
	wndHandler:GetParent():FindChild("RuneOptionFilterUsableContainer"):FindChild("RuneOptionFilterUsableBtn"):SetCheck(false)
	wndHandler:GetParent():FindChild("RuneOptionFilterNoGroupsContainer"):FindChild("RuneOptionFilterNoGroupsBtn"):SetCheck(false)
	wndHandler:GetParent():FindChild("RuneOptionFilterBuffBtn"):SetText("--")

	self.wndMain:FindChild("RuneCreationItemList"):SetText("Select a rune element at the top to view the runes.")
	self.wndMain:FindChild("RuneCreationItemList"):DestroyChildren()
	self.wndMain:FindChild("RuneCreationItemList"):RecalculateContentExtents()
	self:ResetRuneCreationBottom()
	self:RedrawAll()
end

function WhatTheRune:OnRuneOptionFilterBuffBtn( wndHandler, wndControl, eMouseButton )
	local wndFilterlist = self.wndMain:FindChild("RuneOptionBuffList")
	
	wndFilterlist:SetWindowSubclass(wndBBarr)
	self:RedrawBuffFilter()
end

function WhatTheRune:RedrawBuffFilter()
	local wndFilterlist = self.wndMain:FindChild("RuneOptionBuffList")
	wndFilterlist:DestroyChildren()
	
	for idx, t in ipairs(ktAttributeFromId) do
    	local wndItm = self:LoadByName("BuffItem", wndFilterlist, "BuffItem"..idx)
		wndItm:FindChild("BuffItemBtn"):SetData(idx)
		wndItm:FindChild("BuffItemText"):SetText(ktAttributeToText[t])
		
		if self.tUserSettings.Buff == idx then
			wndItm:FindChild("BuffItemBtn"):SetCheck(true)
		end
	end
	
	wndFilterlist:Show(true)

	wndFilterlist:ArrangeChildrenVert(0)
	wndFilterlist:RecalculateContentExtents()
end

---------------------------------------------------------------------------------------------------
-- BuffItem Functions
---------------------------------------------------------------------------------------------------
function WhatTheRune:OnBuffItemCheck( wndHandler, wndControl, eMouseButton )
	local buffData = wndHandler:GetData()
	self.tUserSettings.Buff = buffData 
	
	local wndFilterBtn = self.wndMain:FindChild("RuneOptionFilterBuffBtn")
	wndFilterBtn:SetText(ktAttributeToText[ktAttributeFromId[self.tUserSettings.Buff]])
	local wndFilterlist = self.wndMain:FindChild("RuneOptionBuffList")
	wndFilterlist:Show(false)
end

-----------------------------------------------------------------------------------------------
-- Main Draw Method
-----------------------------------------------------------------------------------------------

function WhatTheRune:RedrawAll()
	if not self.wndMain or not self.wndMain:IsValid() or not self.wndMain:IsVisible() then
		return
	end

	if self.wndMain:FindChild("ToggleRuneCreation"):IsChecked() then
		self.wndMain:FindChild("ToggleRuneOption"):Enable(true)
		self.wndMain:FindChild("ToggleRuneOption"):SetCheck(self.tUserSettings.FiltersShown)
		self.wndMain:FindChild("RuneOptionsContainer"):Show(self.wndMain:FindChild("ToggleRuneOption"):IsChecked())

		self:RedrawCreation()
	elseif self.wndMain:FindChild("ToggleEquipRunes"):IsChecked() then
		self.wndMain:FindChild("ToggleRuneOption"):Enable(false)
		self.wndMain:FindChild("RuneOptionsContainer"):Show(false)

		self:RedrawEquipment()
	end
end

function WhatTheRune:RedrawEquipment()
	local wndParent = self.wndMain:FindChild("MainItemList")
	local nVScrollPos = wndParent:GetVScrollPos()
	wndParent:DestroyChildren()

	for idx, tCurrItem in pairs(CraftingLib.GetValidGlyphableItems(true, false)) do -- 1st arg is Equipped
		if tCurrItem then
			self:NewEquipmentItem(tCurrItem, wndParent, true)
		end
	end

	for idx, tCurrItem in pairs(CraftingLib.GetValidGlyphableItems(false, true)) do -- 2nd arg is Inventory
		if tCurrItem then
			self:NewEquipmentItem(tCurrItem, wndParent)
		end
	end

	wndParent:ArrangeChildrenVert(0)
	wndParent:SetVScrollPos(nVScrollPos)
	wndParent:RecalculateContentExtents()
	wndParent:SetText(#self.wndMain:FindChild("MainItemList"):GetChildren() == 0 and Apollo.GetString("EngravingStation_NoInscribableItems") or "")
end

function WhatTheRune:RedrawCreation()
	-- Determine element filter
	local eElementFilter
	for idx, wndCurr in pairs(self.wndMain:FindChild("RuneCreationElementList"):GetChildren()) do
		local wndCurrBtn = wndCurr:FindChild("RuneCreationElementBtn")
		wndCurrBtn:FindChild("RuneCreationElementIcon"):SetSprite(karSigilElementsToSprite[wndCurrBtn:GetData()].strFade2)
		wndCurrBtn:FindChild("RuneCreationElementIconBright"):SetSprite(karSigilElementsToSprite[wndCurrBtn:GetData()].strBright2)


        -- If loaded with external call (from eg. RuneMaster) check the sigilId to enable button
        if not isCalledExternal then
            wndCurrBtn:FindChild("RuneCreationElementIconBright"):Show( wndCurrBtn:IsChecked(), false, 0.4)
        else
            if karSigilElementsToId[wndCurrBtn:GetData()] == externalElement then
                wndCurrBtn:FindChild("RuneCreationElementIconBright"):Show( true, false, 0.4)
                wndCurrBtn:SetCheck(true)
            else
                wndCurrBtn:FindChild("RuneCreationElementIconBright"):Show( false, false, 0.4)
                wndCurrBtn:SetCheck(false)
            end
        end

        if wndCurrBtn and wndCurrBtn:IsChecked() then
            eElementFilter = karSigilElementsToId[wndCurrBtn:GetData()] or nil
            self.wndMain:FindChild("RuneCreationContainer"):FindChild("SigilName"):SetText(karSigilElementsToSprite[wndCurrBtn:GetData()].strName)
        end
    end

	if not eElementFilter then
		self.wndMain:FindChild("RuneCreationContainer"):FindChild("SigilName"):SetText("--")
		return
	end
	
	-- Prebuild the table
	local tSchematicList = {}
	for idx, tCurrSchematic in pairs(CraftingLib.GetSchematicList(CraftingLib.CodeEnumTradeskill.Runecrafting, eElementFilter, nil, false)) do
		local tSchematicInfo = CraftingLib.GetSchematicInfo(tCurrSchematic.nSchematicId)
		local strItemType = tSchematicInfo.itemOutput:GetItemTypeName() -- This will be Earth Tier 1, Earth Tier 2, etc.
		if not tSchematicList[strItemType] then
			tSchematicList[strItemType] = {}
		end
		tSchematicList[strItemType][idx] = { tCurrSchematic, tSchematicInfo }
	end

	-- Draw Runes
	local wndCreateARuneParent = self.wndMain:FindChild("RuneCreationItemList")
	local tRuneListAll = {}
	for strItemType, tRuneList in pairs(tSchematicList) do
		if not self.tUserSettings.NoGroups then
			local wndTypeTier = self:LoadByName("TypeTierItem", wndCreateARuneParent, "TypeTierItem"..strItemType)
			wndTypeTier:FindChild("TypeTierItemLabel"):SetText(strItemType)
			wndTypeTier:FindChild("TypeTierItemBtn"):SetData(wndTypeTier)
	        wndTypeTier:FindChild("TypeTierItemBtn"):SetCheck(false)
			wndTypeTier:SetData(tRuneList)
		else
			for idx2, tPackedData in pairs(tRuneList) do
				local tCurrSchematic = tPackedData[1]
				local tSchematicInfo = tPackedData[2]
	
				tRuneListAll[idx2] = tPackedData
			end
		end
	end

	if self.tUserSettings.NoGroups then
		local wndTypeTier = self:LoadByName("TypeTierItem", wndCreateARuneParent, "TypeTierItemAll")
		wndTypeTier:FindChild("TypeTierItemLabel"):SetText("All Runetypes...")
		wndTypeTier:FindChild("TypeTierItemBtn"):SetData(wndTypeTier)
		wndTypeTier:FindChild("TypeTierItemBtn"):SetCheck(false)
		wndTypeTier:SetData(tRuneListAll)
	end

	-- Check the first in list, after sorting
	wndCreateARuneParent:SetText("")
	wndCreateARuneParent:ArrangeChildrenVert(0, function(a,b) return a:GetName() < b:GetName() end)

    local wndFirst

    if isCalledExternal then
        wndFirst = wndCreateARuneParent:FindChild("TypeTierItem"..externalTypeTier)
    else
        wndFirst = wndCreateARuneParent:GetChildren()[1]
    end
	
	if self.tUserSettings.NoGroups then
		wndFirst = wndCreateARuneParent:FindChild("TypeTierItemAll")
	end
	
    if wndFirst then
		wndFirst:FindChild("TypeTierItemBtn"):SetCheck(true)
		self:DrawCreateRuneContents(wndFirst)
	end

	self:ResizeCreateRunes()
end

function WhatTheRune:DrawCreateRuneContents(wndParent)
	local tRuneList = wndParent:GetData()

	for idx2, tPackedData in pairs(tRuneList) do
		local tCurrSchematic = tPackedData[1]
		local tSchematicInfo = tPackedData[2]
		local tItemGlyphInfo = tSchematicInfo.itemOutput:GetGlyphInfo()
		
		-- Check Materials
		local bHasMaterials = true
		local iMissingCount = 0;
		for idx3, tMaterialData in pairs(tSchematicInfo.tMaterials) do
			if tMaterialData.nAmount > tMaterialData.itemMaterial:GetBackpackCount() then
				bHasMaterials = false
				iMissingCount = iMissingCount + (tMaterialData.nAmount - tMaterialData.itemMaterial:GetBackpackCount())
			end
		end
		
		local bDisplayRune = true
		
		-- Do we have filters, if so then apply them accordingly
		if self:HelperHaveRuneFilter() == true and not isCalledExternal then
			if self:HelperHaveRuneFilter("Craftable") and not bHasMaterials then
				bDisplayRune = false
			end
			
			if self:HelperHaveRuneFilter("Usable") then
				for idx, tCurrItem in pairs(CraftingLib.GetValidGlyphableItems(true, false)) do -- 1st arg is Equipped
					if tCurrItem then
						if not self:HelperPrereqFailed(tCurrItem) then
							bDisplayRune = false
						end
					end
				end
	
				for idx, tCurrItem in pairs(CraftingLib.GetValidGlyphableItems(false, true)) do -- 2nd arg is Inventory
					if tCurrItem then
						if not self:HelperPrereqFailed(tCurrItem) then
							bDisplayRune = false
						end
					end
				end
			end
			
			if self:HelperHaveRuneFilter("Buff") then
				local tData = tSchematicInfo.itemOutput:GetDetailedInfo()
				local sSearch = string.gsub(ktAttributeToText[ktAttributeFromId[self.tUserSettings.Buff]], " Rating", "")
				if string.find(tData.tPrimary.strFlavor, sSearch ) == nil then
					bDisplayRune = false
				end
			end
		end
		
		
		if bDisplayRune then
			-- Formatting
			local wndCurr = self:LoadByName("NewRuneItem", wndParent:FindChild("TypeTierItemContainer"), "NewRuneItem"..tCurrSchematic.nSchematicId)
			wndCurr:FindChild("NewRuneMoreInfoBtn"):SetData(tSchematicInfo)
			wndCurr:FindChild("NewRuneMoreInfoName"):SetData(bHasMaterials)
	
			local myIcon = wndCurr:FindChild("NewRuneMoreInfoName"):FindChild("NewRuneMoreInfoClassIcon")
			local myClass = tSchematicInfo.itemOutput:GetRequiredClass()
	
			if myClass[1] ~= nil then
				local eClassId = myClass[1].idClassReq
				if eClassId == GameLib.CodeEnumClass.Engineer then
					myIcon:SetSprite("Icon_Windows_UI_CRB_Engineer")
				elseif eClassId  == GameLib.CodeEnumClass.Esper then
					myIcon:SetSprite("Icon_Windows_UI_CRB_Esper")
				elseif eClassId  == GameLib.CodeEnumClass.Spellslinger then
					myIcon:SetSprite("Icon_Windows_UI_CRB_Spellslinger")
				elseif eClassId  == GameLib.CodeEnumClass.Medic then
					myIcon:SetSprite("Icon_Windows_UI_CRB_Medic")
				elseif eClassId  == GameLib.CodeEnumClass.Warrior then
					myIcon:SetSprite("Icon_Windows_UI_CRB_Warrior")
				elseif eClassId == GameLib.CodeEnumClass.Stalker then
					myIcon:SetSprite("Icon_Windows_UI_CRB_Stalker")
				else
					myIcon:SetSprite(nil)
				end
				
				if myClass[1].idClassReq ~= GameLib.GetPlayerUnit():GetClassId() then
					myIcon:SetBGColor("AddonError")
				end
				myIcon:Show(true,true)
			end
			
			wndCurr:FindChild("NewRuneMoreInfoName"):SetText(tCurrSchematic.strName)
	
			if iMissingCount > 0 then
				local wndItem = wndCurr:FindChild("NewRuneMoreInfoName")
				wndItem:FindChild("NewRuneMoreInfoMissingCount"):SetData(tSchematicInfo)
				wndItem:FindChild("NewRuneMoreInfoMissingCount"):SetText("(Missing "..iMissingCount.." item"..(iMissingCount > 1 and "s" or "")..")")
			else
				wndCurr:FindChild("NewRuneMoreInfoName"):FindChild("NewRuneMoreInfoMissingCount"):SetData(nil)		
				wndCurr:FindChild("NewRuneMoreInfoName"):FindChild("NewRuneMoreInfoMissingCount"):SetText("")
			end
			wndCurr:FindChild("NewRuneMoreInfoName"):SetTextColor(bHasMaterials and "UI_BtnTextBlueNormal" or "UI_BtnTextRedNormal")
			wndCurr:FindChild("NewRuneMoreInfoIcon"):SetSprite(bHasMaterials and "sprRunecrafting_BulletPoint_Green" or "sprRunecrafting_BulletPoint_Red")
			Tooltip.GetItemTooltipForm(self, wndCurr, tSchematicInfo.itemOutput, {bSelling = false})
	    end
	end
    if isCalledExternal then
		local wndSelectedRuneItem = self.wndMain:FindChild("NewRuneItem"..externalSchematicId)
	    self:OnExternalHelper(wndSelectedRuneItem)
    end
end

function WhatTheRune:OnTypeTierItemCheck(wndHandler, wndControl)
	local wndParent = wndHandler:GetData()
	self:DrawCreateRuneContents(wndParent)
	self:ResizeCreateRunes()
end

function WhatTheRune:OnTypeTierItemUncheck(wndHandler, wndControl)
	local wndParent = wndHandler:GetData()
	wndParent:FindChild("TypeTierItemContainer"):DestroyChildren()
	self:ResizeCreateRunes()
end

function WhatTheRune:ResizeCreateRunes()
	local function SortTypeTierItemContainer(a,b)
		if not a or not b or not a:FindChild("NewRuneMoreInfoName") or not b:FindChild("NewRuneMoreInfoName") then return true end
		return a:FindChild("NewRuneMoreInfoName"):GetText() < b:FindChild("NewRuneMoreInfoName"):GetText()
	end

	for idx, wndCurr in pairs(self.wndMain:FindChild("RuneCreationItemList"):GetChildren()) do
		local nHeight = 0
		if wndCurr:FindChild("TypeTierItemBtn"):IsChecked() then
			nHeight = wndCurr:FindChild("TypeTierItemContainer"):ArrangeChildrenVert(0, SortTypeTierItemContainer)
		else
			nHeight = wndCurr:FindChild("TypeTierItemContainer"):ArrangeChildrenVert(0)
		end

		local nExtra = wndCurr:FindChild("TypeTierItemBtn"):IsChecked() and 3 or 0
		local nLeft, nTop, nRight, nBottom = wndCurr:GetAnchorOffsets()
		wndCurr:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nHeight + nExtra + self.knDefaultTypeTierItemHeight)
	end
	self.wndMain:FindChild("RuneCreationItemList"):ArrangeChildrenVert(0, function(a,b) return a:GetName() < b:GetName() end)
	
	if isCalledExternal then
		local nRuneCount = 0
		for idx, wndCurr in pairs(self.wndMain:FindChild("RuneCreationItemList"):FindChild("TypeTierItem"..self:HelperIIF(self:HelperHaveRuneFilter("NoGroups"), "All", externalTypeTier)):FindChild("TypeTierItemContainer"):GetChildren()) do
			if wndCurr == self.wndMain:FindChild("NewRuneItem"..externalSchematicId) then
				break
			end
			nRuneCount = nRuneCount + 1
		end	

		self.wndMain:FindChild("RuneCreationItemList"):SetVScrollPos(nRuneCount * 32)
		self.wndMain:FindChild("RuneCreationItemList"):RecalculateContentExtents()
		isCalledExternal = false
	end
end

-----------------------------------------------------------------------------------------------
-- Rune Creation Items
-----------------------------------------------------------------------------------------------

function WhatTheRune:OnUpdateInventory()
	self:HelperRefreshBottomMaterials()
	self:RedrawAll()
end


function WhatTheRune:OnNewRuneMoreInfoCheck(wndHandler, wndControl) -- NewRuneMoreInfoBtn, data is tSchematicInfo
	if wndHandler ~= wndControl then
		return
	end

	wndHandler:FindChild("NewRuneMoreInfoName"):SetTextColor(ApolloColor.new("UI_BtnTextBluePressed"))

	local tSchematicInfo = wndHandler:GetData()
	local wndParent = self.wndMain:FindChild("RuneCreationContainer"):FindChild("RuneCreationBottom")
	wndParent:FindChild("RuneCreationMaterialsList"):DestroyChildren()

    local bNotEnough = false
    for idx2, tData in pairs(tSchematicInfo.tMaterials) do
        local itemMaterial = tData.itemMaterial
        if itemMaterial:GetBackpackCount() < tData.nAmount then
            bNotEnough = true
        end
    end

    wndParent:FindChild("RuneCreationCraftBtn"):Show(self:HelperWhenAtStation(true, false))
    wndParent:FindChild("RuneCreationCraftBtn"):SetData(self:HelperWhenAtStation(tSchematicInfo.nSchematicId, nil))

	if isCalledExternal then
		if self:HooksHelperCanUse( externalAddon, "CraftRune" ) then
		    wndParent:FindChild("RuneCreationCraftBtn"):Enable(not bNotEnough)
		else
			wndParent:FindChild("RuneCreationCraftBtn"):Enable(false)
		end
	else
		wndParent:FindChild("RuneCreationCraftBtn"):Enable(not bNotEnough)
	end
		
	wndParent:FindChild("RuneCreationName"):SetText(tSchematicInfo.strName)
	wndParent:FindChild("RuneCreationMaterialsContainer"):Show(true)
	wndParent:SetData(tSchematicInfo)

	self:HelperRefreshBottomMaterials()
end

function WhatTheRune:OnNewRuneMoreInfoUncheck(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return
	end
	wndHandler:FindChild("NewRuneMoreInfoName"):SetTextColor(wndHandler:FindChild("NewRuneMoreInfoName"):GetData() and "UI_BtnTextBlueNormal" or "UI_BtnTextRedNormal")
	self:ResetRuneCreationBottom()
end

function WhatTheRune:ResetRuneCreationBottom()
	local wndParent = self.wndMain:FindChild("RuneCreationContainer"):FindChild("RuneCreationBottom")
	wndParent:FindChild("RuneCreationMaterialsList"):DestroyChildren()
	wndParent:FindChild("RuneCreationMaterialsContainer"):Show(false)
	wndParent:FindChild("RuneCreationCraftBtn"):Enable(false)
    wndParent:FindChild("RuneCreationCraftBtn"):Show(self:HelperWhenAtStation(true, false))
	wndParent:FindChild("RuneCreationName"):SetText("")
	wndParent:SetData(nil)
end

function WhatTheRune:OnRuneCreationCraftBtn(wndHandler, wndControl)
	CraftingLib.CraftItem(wndHandler:GetData())
end

-----------------------------------------------------------------------------------------------
-- Equipment
-----------------------------------------------------------------------------------------------

function WhatTheRune:NewEquipmentItem(itemSource, wndParent, bEquipped)
	local wndItem = Apollo.LoadForm(self.xmlDoc, "EquipmentItem", wndParent, self)
	wndItem:FindChild("EquipmentItemCantUseIcon"):Show(self:HelperPrereqFailed(itemSource))
	wndItem:FindChild("EquipmentItemIcon"):SetSprite(itemSource:GetIcon())
	--wndItem:FindChild("EquipmentItemSigilContainer"):
	
	if bEquipped then
		Tooltip.GetItemTooltipForm(self, wndItem:FindChild("EquipmentItemIcon"), itemSource, { bPrimary = true, bSelling = false })
		wndItem:FindChild("EquipmentItemName"):SetText(String_GetWeaselString(Apollo.GetString("Runecrafting_EquippedPrefix"), itemSource:GetName()))
	else
		local itemEquipped = itemSource and itemSource:GetEquippedItemForItemType() or nil
		Tooltip.GetItemTooltipForm(self, wndItem:FindChild("EquipmentItemIcon"), itemSource, { bPrimary = true, bSelling = false, itemCompare = itemEquipped })
		wndItem:FindChild("EquipmentItemName"):SetText(itemSource:GetName())
	end

	-- Sigils
	wndItem:FindChild("EquipmentItemSigilContainer"):DestroyChildren()
	local tSigilData = itemSource:GetSigils()
	if tSigilData and tSigilData.bIsDefined then
		local bFirstLockedSlot = true
		for idx, tCurrSigil in pairs(tSigilData.arSigils) do
			local wndSigil = self:LoadByName(tCurrSigil.bUnlocked and "SigilSlot" or "SigilLockedSlot", wndItem:FindChild("EquipmentItemSigilContainer"), "SigilSlot"..idx)
			wndSigil:FindChild("SigilSlotBtn"):SetData({ idx, tCurrSigil, itemSource })
			wndSigil:FindChild("SigilSlotType"):SetSprite(karSigilElementsToSprite[tCurrSigil.eType].strFade)

			-- This is how "strong" a socket is. Current design has this diminish the more slots there are (so an 8 slot drop isn't significantly better than a 7 slot drop)
			local tDetailedInfo = itemSource:GetDetailedInfo()
			if tDetailedInfo and tDetailedInfo.tPrimary.tSigils then
				local tRelevantSlotBudgetData = tDetailedInfo.tPrimary.tSigils.arSigils[idx]
				local strRelevantSlotBudgePercent = math.floor(tRelevantSlotBudgetData.nPercent * 100).."%"
				wndSigil:FindChild("SigilSlotEfficiency"):SetText(strRelevantSlotBudgePercent == "100%" and "Max" or strRelevantSlotBudgePercent)
				wndSigil:FindChild("SigilSlotEfficiency"):SetTextColor((tCurrSigil.bUnlocked or bFirstLockedSlot) and ApolloColor.new("UI_TextHoloBody") or ApolloColor.new("UI_TextMetalBody"))
			end

			-- Different formatting and text for different button states
			local tSigilItemData = Item.GetDataFromId(tCurrSigil.idGlyph)
			if tSigilItemData and wndSigil:FindChild("SigilSlotIcon") then
				wndSigil:FindChild("SigilSlotIcon"):SetSprite(tSigilItemData:GetIcon())
				self:HelperBuildItemTooltip(wndSigil, tSigilItemData)
			elseif tCurrSigil.bUnlocked then
				wndSigil:SetTooltip(String_GetWeaselString(Apollo.GetString("EngravingStation_AvailableSlot"), tCurrSigil.strName))
			elseif bFirstLockedSlot then
				bFirstLockedSlot = false
				wndSigil:FindChild("SigilFakeLockIcon"):Show(true)
				wndSigil:SetTooltip(String_GetWeaselString(Apollo.GetString("EngravingStation_LockedSlot"), tCurrSigil.strName))
			else
				wndSigil:FindChild("SigilSlotBtn"):Enable(false)
				wndSigil:SetTooltip(Apollo.GetString("EngravingStation_LockedChildSlot"))
			end
		end
	end
	
	if isDebugging then
		if tSigilData and tSigilData.bIsDefined then
			local bFirstLockedSlot = true
			for idx, tCurrSigil in pairs(tSigilData.arSigils) do
				local wndSigil2 = self:LoadByName(tCurrSigil.bUnlocked and "SigilSlot" or "SigilLockedSlot", wndItem:FindChild("EquipmentItemSigilContainer"), "SigilSlot2"..idx)
				wndSigil2:FindChild("SigilSlotBtn"):SetData({ idx, tCurrSigil, itemSource })
				wndSigil2:FindChild("SigilSlotType"):SetSprite(karSigilElementsToSprite[tCurrSigil.eType].strFade)
	
				-- This is how "strong" a socket is. Current design has this diminish the more slots there are (so an 8 slot drop isn't significantly better than a 7 slot drop)
				local tDetailedInfo = itemSource:GetDetailedInfo()
				if tDetailedInfo and tDetailedInfo.tPrimary.tSigils then
					local tRelevantSlotBudgetData = tDetailedInfo.tPrimary.tSigils.arSigils[idx]
					local strRelevantSlotBudgePercent = math.floor(tRelevantSlotBudgetData.nPercent * 100).."%"
					wndSigil2:FindChild("SigilSlotEfficiency"):SetText(strRelevantSlotBudgePercent == "100%" and "Max" or strRelevantSlotBudgePercent)
					wndSigil2:FindChild("SigilSlotEfficiency"):SetTextColor((tCurrSigil.bUnlocked or bFirstLockedSlot) and ApolloColor.new("UI_TextHoloBody") or ApolloColor.new("UI_TextMetalBody"))
				end
	
				-- Different formatting and text for different button states
				local tSigilItemData = Item.GetDataFromId(tCurrSigil.idGlyph)
				if tSigilItemData and wndSigil2:FindChild("SigilSlotIcon") then
					wndSigil2:FindChild("SigilSlotIcon"):SetSprite(tSigilItemData:GetIcon())
					self:HelperBuildItemTooltip(wndSigil2, tSigilItemData)
				elseif tCurrSigil.bUnlocked then
					wndSigil2:SetTooltip(String_GetWeaselString(Apollo.GetString("EngravingStation_AvailableSlot"), tCurrSigil.strName))
				elseif bFirstLockedSlot then
					bFirstLockedSlot = false
					wndSigil2:FindChild("SigilFakeLockIcon"):Show(true)
					wndSigil2:SetTooltip(String_GetWeaselString(Apollo.GetString("EngravingStation_LockedSlot"), tCurrSigil.strName))
				else
					wndSigil2:FindChild("SigilSlotBtn"):Enable(false)
					wndSigil2:SetTooltip(Apollo.GetString("EngravingStation_LockedChildSlot"))
				end
			end
		end
	end
	
	wndItem:FindChild("EquipmentItemSigilContainer"):ArrangeChildrenHorz(0)
end

function WhatTheRune:OnUnlockConfirmYes(wndHandler, wndControl)
	local nSlotIndex = wndHandler:GetData()[1]
	local itemSource = wndHandler:GetData()[3] -- Data is passed along, origates from SigilSlotBtn and is { idx, tCurrSigil, itemSource }
	CraftingLib.UnlockSigil(itemSource, nSlotIndex)
	self:OnCurrentConfirmPopupClose()
end

function WhatTheRune:OnClearConfirmYes(wndHandler, wndControl)
	local nSlotIndex = wndHandler:GetData()[1]
	local itemSource = wndHandler:GetData()[3] -- Data is passed along, origates from SigilSlotBtn and is { idx, tCurrSigil, itemSource }
	CraftingLib.ClearSigil(itemSource, nSlotIndex)
	self:OnCurrentConfirmPopupClose()
end

-----------------------------------------------------------------------------------------------
-- wndCurrentConfirmPopup
-----------------------------------------------------------------------------------------------

function WhatTheRune:OnSigilLockedSlotBtn(wndHandler, wndControl) -- SigilSlotBtn of SigilLockedSlot
	if not self.bAllowClicks then
		return
	end
	
	local nSlotIndex = wndHandler:GetData()[1]
	local itemSource = wndHandler:GetData()[3]
	local tEngravingInfo = CraftingLib.GetEngravingInfo(itemSource)
	local tCurrentCosts = tEngravingInfo.tUnlockInfo.tCost

	self:OnCurrentConfirmPopupClose()
	self.wndCurrentConfirmPopup = Apollo.LoadForm(self.xmlDoc, "UnlockConfirm", self.wndMain, self)
	self.wndCurrentConfirmPopup:FindChild("UnlockConfirmYes"):SetData(wndHandler:GetData())
	self.wndCurrentConfirmPopup:FindChild("UnlockCashWindow"):SetAmount(tCurrentCosts and tCurrentCosts.monCost or 0)
	self:HelperResizeConfirmPopup(wndHandler)
end

function WhatTheRune:OnSigilSlotBtn(wndHandler, wndControl) -- SigilSlotBtn of SigilLockedSlot
	if not self.bAllowClicks then
		return
	end
	
	local nSlotIndex = wndHandler:GetData()[1]
	local tCurrSigil = wndHandler:GetData()[2]
	local itemSource = wndHandler:GetData()[3]

	self:OnCurrentConfirmPopupClose()

	if self.itemDragging and self.itemDragging:GetGlyphInfo().eType ~= tCurrSigil.eType then
		return Apollo.DragDropQueryResult.Accept -- HACK
	end

	if tCurrSigil.idGlyph == 0 then
		self.wndCurrentConfirmPopup = Apollo.LoadForm(self.xmlDoc, "AddPicker", self.wndMain, self)
		for idx, itemGlyph in pairs(CraftingLib.GetValidGlyphItems()) do
			local tItemGlyphInfo = itemGlyph:GetGlyphInfo()
			if self:HelperDoSigilAndGlyphMatch(tItemGlyphInfo, tCurrSigil) then
				local wndGlyph = self:LoadByName("GlyphPickerItem", self.wndCurrentConfirmPopup:FindChild("AddPickerList"), "AddPickerList"..idx..itemGlyph:GetName())
				wndGlyph:FindChild("GlyphPickerBtn"):SetData({wndHandler, itemGlyph})
				wndGlyph:FindChild("GlyphPickerText"):SetText(itemGlyph:GetName())
				wndGlyph:FindChild("GlyphPickerIcon"):SetSprite(itemGlyph:GetIcon())
				wndGlyph:FindChild("GlyphPickerIcon"):SetText(itemGlyph:GetBackpackCount())
				self:HelperBuildItemTooltip(wndGlyph, itemGlyph)
			end
		end
		local bHasItems = #self.wndCurrentConfirmPopup:FindChild("AddPickerList"):GetChildren() > 0
		self.wndCurrentConfirmPopup:FindChild("AddPickerList"):ArrangeChildrenTiles(0)
		self.wndCurrentConfirmPopup:FindChild("AddPickerList"):SetText(bHasItems and "" or Apollo.GetString("EngravingStation_NoRunesFound"))
	else
		local tEngravingInfo = CraftingLib.GetEngravingInfo(itemSource)
		local tCurrentCosts = tEngravingInfo.tClearInfo.arCosts[nSlotIndex]
		self.wndCurrentConfirmPopup = Apollo.LoadForm(self.xmlDoc, "ClearConfirm", self.wndMain, self)
		self.wndCurrentConfirmPopup:FindChild("ClearConfirmYes"):SetData(wndHandler:GetData())
		self.wndCurrentConfirmPopup:FindChild("ClearCashWindow"):SetAmount(tCurrentCosts and tCurrentCosts.monCost or 0)
	end
	self:HelperResizeConfirmPopup(wndHandler)

	return Apollo.DragDropQueryResult.Accept -- HACK
end

function WhatTheRune:OnGlyphPickerBtn(wndHandler, wndControl)
	self:DrawGlyphAddConfirm(wndHandler:GetData()[1], wndHandler:GetData()[2])
end

function WhatTheRune:DrawGlyphAddConfirm(wndHandler, itemGlyph)
	local nSlotIndex = wndHandler:FindChild("SigilSlotBtn"):GetData()[1]
	local itemSource = wndHandler:FindChild("SigilSlotBtn"):GetData()[3]
	self:OnCurrentConfirmPopupClose()

	-- What the glyph will add to the item, if successful
	local tItemEffectData = itemSource:GetGlyphBonus(itemGlyph, nSlotIndex)
	local strItemEffect = String_GetWeaselString(Apollo.GetString("EngravingStation_EffectIfSuccessful"), tItemEffectData.nValue or 0, ktAttributeToText[tItemEffectData.eProperty])

	-- Draw Window
	self.wndCurrentConfirmPopup = Apollo.LoadForm(self.xmlDoc, "AddConfirm", self.wndMain, self)
	self.wndCurrentConfirmPopup:FindChild("AddConfirmYes"):SetData({ nSlotIndex, itemGlyph, itemSource }) -- GOTCHA: Different 2nd argument than normal
	self.wndCurrentConfirmPopup:FindChild("AddConfirmItemEffect"):SetText(strItemEffect)
	self.wndCurrentConfirmPopup:FindChild("AddConfirmIcon"):SetSprite(itemGlyph:GetIcon())
	self.wndCurrentConfirmPopup:FindChild("AddConfirmText"):Show(itemSource:IsSoulbound())
	self.wndCurrentConfirmPopup:FindChild("AddConfirmSoulboundWarning"):Show(not itemSource:IsSoulbound())
	self:HelperBuildItemTooltip(self.wndCurrentConfirmPopup:FindChild("AddConfirmIcon"), itemGlyph)
	self:HelperResizeConfirmPopup(wndHandler)
end

function WhatTheRune:DrawGlyphReplaceConfirm(wndHandler, itemGlyph)
	local nSlotIndex = wndHandler:FindChild("SigilSlotBtn"):GetData()[1]
	local itemSource = wndHandler:FindChild("SigilSlotBtn"):GetData()[3]
	local tEngravingInfo = CraftingLib.GetEngravingInfo(itemSource)
	local tCurrentCosts = tEngravingInfo.tClearInfo.arCosts[nSlotIndex]
	self:OnCurrentConfirmPopupClose()

	-- Draw Window
	self.wndCurrentConfirmPopup = Apollo.LoadForm(self.xmlDoc, "ReplaceConfirm", self.wndMain, self)
	self.wndCurrentConfirmPopup:FindChild("ReplaceConfirmYes"):SetData({ nSlotIndex, itemGlyph, itemSource }) -- GOTCHA: Different 2nd argument than normal
	self.wndCurrentConfirmPopup:FindChild("ReplaceConfirmIcon"):SetSprite(itemGlyph:GetIcon())
	self.wndCurrentConfirmPopup:FindChild("ReplaceCashWindow"):SetAmount(tCurrentCosts and tCurrentCosts.monCost or 0)
	self:HelperBuildItemTooltip(self.wndCurrentConfirmPopup:FindChild("ReplaceConfirmIcon"), itemGlyph)
	self:HelperResizeConfirmPopup(wndHandler)
end

function WhatTheRune:OnCurrentConfirmPopupClose(wndHandler, wndControl)
	if self.wndCurrentConfirmPopup and self.wndCurrentConfirmPopup:IsValid() then
		self.wndCurrentConfirmPopup:Destroy()
		self.wndCurrentConfirmPopup = nil
		Apollo.StartTimer("WhatTheRune_ConfirmPopupDelay")
		self.bAllowClicks = false
	end
end

function WhatTheRune:OnWhatTheRune_ConfirmPopupDelay()
	self.bAllowClicks = true
end


-----------------------------------------------------------------------------------------------
-- Drag Drop Events
-----------------------------------------------------------------------------------------------

function WhatTheRune:OnSystemBeginDragDrop()
	self.bDragging = true
end

function WhatTheRune:OnSystemEndDragDrop()
	self.bDragging = false
end

function WhatTheRune:OnSigilSlotQueryDragDrop(wndHandler, wndControl, x, y, wndSource, strType, nDragInventoryItemIdx, eResult)
	local itemSource = self.wndMain:FindChild("HiddenBagWindow"):GetItem(nDragInventoryItemIdx)
	if itemSource and strType == "DDBagItem" then
		local tCurrSigil = wndHandler:GetData()[2]
		local tItemGlyphInfo = self.wndMain:FindChild("HiddenBagWindow"):GetItem(nDragInventoryItemIdx):GetGlyphInfo()
		if tCurrSigil.eType and tItemGlyphInfo then
			return Apollo.DragDropQueryResult.Accept
		end
	end
	return Apollo.DragDropQueryResult.Ignore
end

function WhatTheRune:OnSigilSlotDragDrop(wndHandler, wndControl, x, y, wndSource, strType, nDragInventoryItemIdx, bDragDropHasBeenReset)
	local itemSource = self.wndMain:FindChild("HiddenBagWindow"):GetItem(nDragInventoryItemIdx)
	if itemSource and strType == "DDBagItem" then
		local tCurrSigil = wndHandler:GetData()[2]
		local tItemGlyphInfo = self.wndMain:FindChild("HiddenBagWindow"):GetItem(nDragInventoryItemIdx):GetGlyphInfo()
		if self:HelperDoSigilAndGlyphMatch(tItemGlyphInfo, tCurrSigil) then
			if tCurrSigil.idGlyph == 0 then -- No rune in slot
				self:DrawGlyphAddConfirm(wndHandler:GetParent(), self.wndMain:FindChild("HiddenBagWindow"):GetItem(nDragInventoryItemIdx))
			else -- Replace rune in slot
				self:DrawGlyphReplaceConfirm(wndHandler:GetParent(), self.wndMain:FindChild("HiddenBagWindow"):GetItem(nDragInventoryItemIdx))
			end
		end
		return false
	end
	return true
end

function WhatTheRune:OnAddConfirmYes(wndHandler, wndControl) -- Potentially from drag drop or from picker
	local nSlotIndex = wndHandler:GetData()[1]
	local itemGlyph = wndHandler:GetData()[2] -- GOTCHA: This is an item object, not the normal table of data
	local itemSource = wndHandler:GetData()[3]

	local tSigilData = itemSource:GetSigils()
	if tSigilData and tSigilData.bIsDefined then
		local tListOfGlyphs = {}
		for idx, tCurrSigil in pairs(tSigilData.arSigils) do
			if tCurrSigil.bUnlocked then
				tListOfGlyphs[idx] = tCurrSigil.idGlyph
			end
		end
		tListOfGlyphs[nSlotIndex] = itemGlyph:GetItemId() -- Replace with the desired
		CraftingLib.InstallGlyphs(itemSource, tListOfGlyphs)
	end
	self:OnCurrentConfirmPopupClose()
end

function WhatTheRune:OnReplaceConfirmYes( wndHandler, wndControl, eMouseButton )
	local nSlotIndex = wndHandler:GetData()[1]
	local itemGlyph = wndHandler:GetData()[2] -- GOTCHA: This is an item object, not the normal table of data
	local itemSource = wndHandler:GetData()[3]

	local tSigilData = itemSource:GetSigils()
	if tSigilData and tSigilData.bIsDefined then
		local tListOfGlyphs = {}
		for idx, tCurrSigil in pairs(tSigilData.arSigils) do
			if tCurrSigil.bUnlocked then
				tListOfGlyphs[idx] = tCurrSigil.idGlyph
			end
		end
		tListOfGlyphs[nSlotIndex] = itemGlyph:GetItemId() -- Replace with the desired
		CraftingLib.ClearSigil(itemSource, nSlotIndex)
		CraftingLib.InstallGlyphs(itemSource, tListOfGlyphs)
	end

	self:OnCurrentConfirmPopupClose()
end

-----------------------------------------------------------------------------------------------
-- Helpers
-----------------------------------------------------------------------------------------------


function WhatTheRune:LoadByName(strForm, wndParent, strCustomName)
	local wndNew = wndParent:FindChild(strCustomName)
	if not wndNew then
		wndNew = Apollo.LoadForm(self.xmlDoc, strForm, wndParent, self)
		wndNew:SetName(strCustomName)
	end
	return wndNew
end

---------------------------------------------------------------------------------------------------
-- NewRuneItem Functions
---------------------------------------------------------------------------------------------------

function WhatTheRune:OnMouseEnter( wndHandler, wndControl, x, y )
	if wndHandler ~= wndControl or not wndHandler:ContainsMouse() then
		return
	end

	local tSchematicInfo = wndHandler:GetData()
	if not tSchematicInfo then
		return
	end
	
	MissingMaterialsTooltipForm = self.wndMain:LoadTooltipForm("WhatTheRune.xml", "MissingMaterialsTooltipForm")
	MissingMaterialsTooltipForm:FindChild("MMContainer"):FindChild("MMHeader"):FindChild("MMHeaderText"):SetText(tSchematicInfo.strName)
	
	local wndList = MissingMaterialsTooltipForm:FindChild("MMContainer"):FindChild("MMItemList")
	
	
	local iRows = 0
	for idx2, tData in pairs(tSchematicInfo.tMaterials) do
		local itemMaterial = tData.itemMaterial
		
		if itemMaterial:GetBackpackCount() < tData.nAmount then
			iRows = iRows + 1
			local iMissingCount = tData.nAmount - itemMaterial:GetBackpackCount()
			local wndMaterial = self:LoadByName("MissingMaterialsItem", wndList, "MissingMaterialsItem"..itemMaterial:GetName())
			wndMaterial:FindChild("MMIcon"):SetSprite(itemMaterial:GetIcon())
			wndMaterial:FindChild("MMLabel"):SetText(iMissingCount.."x "..itemMaterial:GetName())
		end
	end
	
	local nLeft, nTop, nRight, nBottom = MissingMaterialsTooltipForm:GetAnchorOffsets()
	MissingMaterialsTooltipForm:SetAnchorOffsets(nLeft, nTop, nRight, nBottom + (iRows * 32))
	
	wndList:ArrangeChildrenVert(0, function(a,b) return a:GetName() < b:GetName() end)

	wndHandler:SetTooltipForm(MissingMaterialsTooltipForm)
	
	tSchematicInfo = nil
end

function WhatTheRune:OnMouseExit( wndHandler, wndControl, x, y )
	wndHandler:SetTooltipForm(nil)
	if MissingMaterialsTooltipForm then
		MissingMaterialsTooltipForm:Destroy()
	end
end


----------------------------------------------------------------------------------------------
-- External hooks so other addons can load WhatTheRune and use the functionality
----------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------
-- WhatTheRune:WhatTheVersion( sAddonName, nVersion, bReturnHooks )
-- Returns on success: bool, nVersion, (optional tHooks)
-- Returns on error: bool, strMsg 
----------------------------------------------------------------------------------------------
function WhatTheRune:WhatTheVersion( sAddonName, nVersion, bReturnHooks )
	if sAddonName == nil or nVersion == nil then
		return false, nil
	end
	
	local bRetval, sMsg = self:HooksHelperCanUse( sAddonName, "Version", nVersion )
	if bRetval then
		externalAddon = sAddonName
		if bReturnHooks == nil then bReturnHooks = false end
		
		if bReturnHooks then
			return true, knAddonVersion, tHookFunctions
		else
			return true, knAddonVersion
		end
	else
		return false, sMsg	
	end
end

----------------------------------------------------------------------------------------------
-- WhatTheRune:OnExternalViewRune( nSchematicId )
-- Returns on success: bool
-- Returns on error: bool, strMsg 
----------------------------------------------------------------------------------------------
function WhatTheRune:OnExternalViewRune( nSchematicId )
	if not self:HooksHelperCanUse( externalAddon, "ViewRune" ) then
		return false, "No rights"
	end
    if not nSchematicId or tonumber(nSchematicId) == nil then
        return false, "No schematicId"
    end

    isCalledExternal = true

    externalSchematicId = tonumber(nSchematicId)

    local tSchematicInfo = CraftingLib.GetSchematicInfo( nSchematicId )

    externalElement = tSchematicInfo.itemOutput:GetItemCategory()
    externalTypeTier = tSchematicInfo.itemOutput:GetItemTypeName()
	
	self:HooksHelperPhysicalAtStation()

	if self.wndMain and self.wndMain:IsValid() then
		if not self.wndMain:FindChild("ToggleRuneCreation"):IsChecked() then
			self.wndMain:FindChild("ToggleRuneCreation"):SetCheck(true)
			self.wndMain:FindChild("ToggleEquipRunes"):SetCheck(false)
		end
		self.wndMain:ToFront()
		self.wndMain:FindChild("RuneCreationItemList"):SetText("Select a rune element at the top to view the runes.")
		self.wndMain:FindChild("RuneCreationItemList"):DestroyChildren()
		self.wndMain:FindChild("RuneCreationItemList"):RecalculateContentExtents()
		self:ResetRuneCreationBottom()
		self:RedrawAll()
	else
	    self:Initialize()
    	Apollo.StartTimer("WhatTheRune_InitializeDelay") -- Will do RedrawAll()
	end
	
	return true
end

----------------------------------------------------------------------------------------------
-- WhatTheRune:OnExternalCraftRune( nSchematicId )
-- Returns on success: bool
-- Returns on error: bool, strMsg 
----------------------------------------------------------------------------------------------
function WhatTheRune:OnExternalCraftRune( nSchematicId )
	if not self:HooksHelperCanUse( externalAddon, "CraftRune" ) then
		return false, "No rights"
	end
    if not nSchematicId or tonumber(nSchematicId) == nil then
        return false, "No schematicId"
    end

    isCalledExternal = true

    externalSchematicId = tonumber(nSchematicId)

    local tSchematicInfo = CraftingLib.GetSchematicInfo( nSchematicId )

    externalElement = tSchematicInfo.itemOutput:GetItemCategory()
    externalTypeTier = tSchematicInfo.itemOutput:GetItemTypeName()
	
	self:HooksHelperPhysicalAtStation()

	if self.wndMain and self.wndMain:IsValid() then
		if not self.wndMain:FindChild("ToggleRuneCreation"):IsChecked() then
			self.wndMain:FindChild("ToggleRuneCreation"):SetCheck(true)
			self.wndMain:FindChild("ToggleEquipRunes"):SetCheck(false)
		end
		self.wndMain:ToFront()
		self.wndMain:FindChild("RuneCreationItemList"):SetText("Select a rune element at the top to view the runes.")
		self.wndMain:FindChild("RuneCreationItemList"):DestroyChildren()
		self.wndMain:FindChild("RuneCreationItemList"):RecalculateContentExtents()
		self:ResetRuneCreationBottom()
		self:RedrawAll()
	else
	    self:Initialize()
    	Apollo.StartTimer("WhatTheRune_InitializeDelay") -- Will do RedrawAll()
	end
	
	return true
end
----------------------------------------------------------------------------------------------
-- WhatTheRune Hooks helpers
-----------------------------------------------------------------------------------------------

function WhatTheRune:OnExternalHelper( wndRuneItem )
	if not self.wndMain or not self.wndMain:IsValid() or not self.wndMain:IsVisible() or not wndRuneItem then
		return
	end

	wndRuneItem:FindChild("NewRuneMoreInfoBtn"):SetCheck(true)
    wndRuneItem:FindChild("NewRuneMoreInfoBtn"):FindChild("NewRuneMoreInfoName"):SetTextColor(ApolloColor.new("UI_BtnTextBluePressed"))
	   
	local tSchematicInfo = wndRuneItem:FindChild("NewRuneMoreInfoBtn"):GetData()
    local wndParent = self.wndMain:FindChild("RuneCreationBottom")

    wndParent:FindChild("RuneCreationMaterialsList"):DestroyChildren()

    local bNotEnough = false
    for idx2, tData in pairs(tSchematicInfo.tMaterials) do
        local itemMaterial = tData.itemMaterial
        if itemMaterial:GetBackpackCount() < tData.nAmount then
            bNotEnough = true
        end
    end
	
	if isCalledExternal and self:HooksHelperCanUse( externalAddon, "CraftRune" ) then
	    wndParent:FindChild("RuneCreationCraftBtn"):Show(self:HelperWhenAtStation(true, false))
	    wndParent:FindChild("RuneCreationCraftBtn"):SetData(self:HelperWhenAtStation(tSchematicInfo.nSchematicId, nil))
	    wndParent:FindChild("RuneCreationCraftBtn"):Enable(not bNotEnough)
	else
		wndParent:FindChild("RuneCreationCraftBtn"):Enable(false)
	end
	
	wndParent:FindChild("RuneCreationName"):SetText(tSchematicInfo.strName)
    wndParent:FindChild("RuneCreationMaterialsContainer"):Show(true)
    wndParent:SetData(tSchematicInfo)

    local tSchematicInfo = wndParent:GetData()
    if not tSchematicInfo then
        return
    end


    for idx2, tData in pairs(tSchematicInfo.tMaterials) do
        local itemMaterial = tData.itemMaterial
        local bNotEnough = itemMaterial:GetBackpackCount() < tData.nAmount
        local wndMaterial = self:LoadByName("RawMaterialsItem", wndParent:FindChild("RuneCreationMaterialsList"), "RawMaterialsItem"..itemMaterial:GetName())
        wndMaterial:FindChild("RawMaterialsIcon"):SetSprite(itemMaterial:GetIcon())
        wndMaterial:FindChild("RawMaterialsIcon"):SetText(String_GetWeaselString(Apollo.GetString("CRB_NOutOfN"), itemMaterial:GetBackpackCount(), tData.nAmount))
        wndMaterial:FindChild("RawMaterialsNotEnough"):Show(bNotEnough)

        Tooltip.GetItemTooltipForm(self, wndMaterial, itemMaterial, {bSelling = false})
    end
    wndParent:FindChild("RuneCreationMaterialsList"):ArrangeChildrenHorz(0)

end

function WhatTheRune:HooksHelperCanUse( sAddonName, sHook, nAddonVersion )
	if sAddonName == nil or sHook == nil then
        return false, "missing params"
    end

    local tKnownAddon = knAddonCompatibilities[sAddonName]

    if tKnownAddon == nil then
        tKnownAddon = knAddonCompatibilities["Unknown"]
    end
	
	if nAddonVersion ~= nil then
	    if tKnownAddon.nVersion > nAddonVersion then
	        return false, "not compatible"
	    end
	end
	
    if tHooks[sHook] > tKnownAddon.nMaxHook then
        return false, "no rights"
    end

    return true
end

function WhatTheRune:HooksHelperPhysicalAtStation()
	if not isCalledExternal then
		isAtEngravingStation = false
		return isAtEngravingStation
	end
	
	-- bugreport, when you open WTR through 3rd party addon and never have opened it directly, it crashes on the LastKnownEngravingStationId, because it is nil.
	if self.tUserSettings.LastKnownEngravingStationId == nil or tonumber(self.tUserSettings.LastKnownEngravingStationId) == nil then
		self.tUserSettings.LastKnownEngravingStationId = 0
	end
	
	GameLib.SetTargetUnit(GetUnit(self.tUserSettings.LastKnownEngravingStationId))
	local uEngravingStation = GameLib.GetTargetUnit()
	local uPlayer = GameLib.GetPlayerUnit()

	if uEngravingStation == nil or uPlayer == nil then
		self:DebugPrint("Not at engraving station")
		isAtEngravingStation = false
		return isAtEngravingStation
	end
	
	local ues = uEngravingStation:GetPosition()
	local up = uPlayer:GetPosition()
	
	if up.x < ues.x - 10 or up.x > ues.x + 10 then
		self:DebugPrint("px:"..up.x..", esx:"..ues.x)
		isAtEngravingStation = false
		return isAtEngravingStation
	end
	if up.y < ues.y - 10 or up.y > ues.y + 10 then
		self:DebugPrint("py:"..up.y..", esy:"..ues.y)
		isAtEngravingStation = false
		return isAtEngravingStation
	end
	if up.z < ues.z - 10 or up.z > ues.z + 10 then
		self:DebugPrint("pz:"..up.z..", eszz:"..ues.z)
		isAtEngravingStation = false
		return isAtEngravingStation
	end
	
	isAtEngravingStation = true
	return isAtEngravingStation
end

function WhatTheRune:DebugPrint(msg)
	if isDebugging then
		Print("wtrDebug: "..msg)
	end
end

----------------------------------------------------------------------------------------------
-- Global helpers
-----------------------------------------------------------------------------------------------

function WhatTheRune:HelperRover( id, var )
	if isDebugging and Rover ~= nil then
		Rover:AddWatch(id, var)
	else
		self:DebugPrint("Rover not found")
	end
end
	
function WhatTheRune:HelperRefreshBottomMaterials()
	if not self.wndMain or not self.wndMain:IsValid() then
		return
	end

	local wndParent = self.wndMain:FindChild("RuneCreationContainer"):FindChild("RuneCreationBottom")
	local tSchematicInfo = wndParent:GetData()
	if not tSchematicInfo then
		return
	end


	for idx2, tData in pairs(tSchematicInfo.tMaterials) do
		local itemMaterial = tData.itemMaterial
		local bNotEnough = itemMaterial:GetBackpackCount() < tData.nAmount
		local wndMaterial = self:LoadByName("RawMaterialsItem", wndParent:FindChild("RuneCreationMaterialsList"), "RawMaterialsItem"..itemMaterial:GetName())
		wndMaterial:FindChild("RawMaterialsIcon"):SetSprite(itemMaterial:GetIcon())
		wndMaterial:FindChild("RawMaterialsIcon"):SetText(String_GetWeaselString(Apollo.GetString("CRB_NOutOfN"), itemMaterial:GetBackpackCount(), tData.nAmount))
		wndMaterial:FindChild("RawMaterialsNotEnough"):Show(bNotEnough)

		Tooltip.GetItemTooltipForm(self, wndMaterial, itemMaterial, {bSelling = false})
	end
	wndParent:FindChild("RuneCreationMaterialsList"):ArrangeChildrenHorz(0)
end

function WhatTheRune:HelperResizeConfirmPopup(wndOrigin)
	local wndIter = wndOrigin
	local wndExit = wndOrigin:GetFrame()
	local nTotalX, nTotalY = 0, 0

	for idx = 1, 99 do
	local nIterTrap = 0
		if not wndIter or wndIter == wndExit then
			break
		end

		local x, y = wndIter:GetPos()
		nTotalX = nTotalX + x
		nTotalY = nTotalY + y
		wndIter = wndIter:GetParent()
	end

	local wndPopup = self.wndCurrentConfirmPopup
	local nLeft, nTop, nRight, nBottom = wndPopup:GetAnchorOffsets()
	wndPopup:SetAnchorOffsets(nTotalX + nLeft, nTotalY + nTop, nTotalX + nRight, nTotalY + nBottom)
	wndPopup:ToFront()
end

function WhatTheRune:HelperDoSigilAndGlyphMatch(tItemGlyphInfo, tSigil)
	return tSigil.eType == Item.CodeEnumSigilType.Omni or tItemGlyphInfo.eType == Item.CodeEnumSigilType.Omni or tItemGlyphInfo.eType == tSigil.eType
end

function WhatTheRune:HelperPrereqFailed(tCurrItem)
	return tCurrItem and tCurrItem:IsEquippable() and not tCurrItem:CanEquip()
end

function WhatTheRune:HelperBuildItemTooltip(wndArg, itemCurr)
	Tooltip.GetItemTooltipForm(self, wndArg, itemCurr, { bPrimary = true, bSelling = false, itemCompare = itemCurr:GetEquippedItemForItemType() })
end

function WhatTheRune:HelperWhenAtStation(objTrue, objFalse)
    if isAtEngravingStation then
        return objTrue
    else
        return objFalse
    end
end

-- Helper for intermediate if (iif), give parameter to check (boolean!) then true result and the false result
function WhatTheRune:HelperIIF(bParam, objTrue, objFalse)
    if bParam then
        return objTrue
    else
        return objFalse
    end
end

function WhatTheRune:HelperHaveRuneFilter( sFilter )
	if sFilter == nil then
		if self.tUserSettings.Craftable or self.tUserSettings.Usable or self.tUserSettings.NoGroups or self.tUserSettings.Buff then
			return true
		end
	else
		if self.tUserSettings[sFilter] == nil then
			return false
		else
			return self.tUserSettings[sFilter] 
		end
	end 
		
	return false
end



----------------------------------------------------------------------------------------------
-- UnitEvents listeners
-----------------------------------------------------------------------------------------------

function WhatTheRune:OnUnitCreated( unitId )
	if unitId == nil then
		return
	end
	
	if unitId:GetName() == Apollo.GetString("EngravingStation_TitleText") then
		if self.tUserSettings.LastKnownEngravingStationId ~= unitId:GetId() then
			self.tUserSettings.LastKnownEngravingStationId = unitId:GetId()
		end
	end
end	
	
function WhatTheRune:OnInteractiveUnitChanged( unitArg, strArg )
	if unitArg == nil then
		return
	end
	
	if unitArg:GetName() == Apollo.GetString("EngravingStation_TitleText") then
		if self.tUserSettings.LastKnownEngravingStationId ~= unitArg:GetId() then
			self.tUserSettings.LastKnownEngravingStationId = unitArg:GetId()
		end
	end
end


----------------------------------------------------------------------------------------------
-- WhatTheRune Instance
-----------------------------------------------------------------------------------------------
local WhatTheRuneInst = WhatTheRune:new()
WhatTheRuneInst:Init()
